/**
 * Copyright 2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 */
package com.sunlabs.archsim.fleet.ships.constant;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.fleet.common.OutputFunction;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 * @author ib27688
 *
 */
public class Constant extends Component {

	protected Data constantValue;
	
	Terminal out;
	Terminal ackOut;
	ConstantAction constantAction;
	
	public Constant(String name, Data c) {
		super(name, null);
		
		// create and assemble the action and terminals
		assemble();
		
		constantValue= c;
	} // constant
	
	
	/**
	 * Constructor - override any assigned value, take the value from
	 * the parameter argument from the XML setup.
	 * @param name = component name
	 * @param params parameters taken from the XML input
	 */
	public Constant(String name, Parameters params) {
		super(name, null);
		
		// use a default value from ConstantDefaults
		// info-log what happened
		constantValue= ConstantDefaults.CONSTANT_VALUE_DEF;
		Delay delayParam= null;

		// process the parameters
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				boolean typeFound= true;
				if (pName.equals(ConstantDefaults.CONSTANT_VALUE) 
						|| pName.equals(FleetGlobal.PARAMETER) ) {
					// find the type for the value. is it an integer?
					if (sValue == null || sValue.equals("null")) {
						constantValue= null;
					}
					else {
						// is it an integer
						try {
							int intValue= Integer.parseInt(sValue);
							constantValue= new Data(intValue);
						}
						catch (Exception e) {
							typeFound= false;
						}
					}
					// not an integer, is it boolean?
					if (!typeFound) {
						try {
							boolean boolValue= Boolean.parseBoolean(sValue);
							constantValue= new Data(boolValue);
						}
						catch (Exception e) {
							typeFound= false;
						}					
					}
					// not a boolean, use the string
					constantValue= new Data(sValue);
				}
				else if (pName.equals(ConstantDefaults.CONSTANT_DELAY) ) {
					try {
						int d= Integer.parseInt(sValue);
						delayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ ConstantDefaults.CONSTANT_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");					
				} // else

				
			} // for	
		} // if
		
		
		// create and assemble the action and terminals
		assemble();
		
		
		// System.out.println("Constant value= " + constantValue);
		
		// check whether delay needs to be set
		if (delayParam != null) {
			constantAction.constFunction.setDelay(delayParam);
		}
	} // Constant constructor
	
	protected void assemble() {
		// make the terminals
		out= new OutputTerminal(FleetGlobal.OUT, this);
		ackOut= new InputTerminal(FleetGlobal.ACK_OUT, this);
		// make the action a
		constantAction= new ConstantAction("ConstantAction");
		addMember(constantAction);
		// attach terminals
		constantAction.attach(out);
		constantAction.attach(ackOut);		
	} // assemble
	
	public Terminal getTerminal(String name) {
		if (name.equals(out.getBaseName()) 
				|| name.equals(out.getBaseAlias()))
			return out;
		else if (name.equals(ackOut.getBaseName()) 
				|| name.equals(ackOut.getBaseAlias()))
			return ackOut;
		else 
			return null;
	} // getTerminal
	
	// used for testing, return the value
	public Data value() {
		return constantValue;
	}
	
	
	protected class ConstantAction extends Action {
		
		public DeliverConstantFn constFunction;
		
		public ConstantAction(String name) {
			super(name);
			constFunction= new DeliverConstantFn();
			outputFunctions.put(FleetGlobal.OUT, constFunction);
		}
		
		private class DeliverConstantFn extends OutputFunction {
			
			public DeliverConstantFn() {
				super("DeliverConstantFunction");
			
			Integer fDelay= globals.intValue(ConstantDefaults.CONSTANT_DELAY);
			functionDelay= (fDelay != null)
				? new Delay(fDelay.intValue())
				: ConstantDefaults.CONSTANT_DELAY_DEF;
			}
			
			public Data value() {
				return (Data)(constantValue.clone());
			}
		}
		
	} // ConstantAction
	
	
} // class Constant
