package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.bind.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** a tree (or node in a tree); see jargon.txt for details */
public class Tree<T>
    extends PrintableTree<Tree<T>>
    implements Iterable<Tree<T>>,
               GraphViz.ToGraphViz {

    private final T            head;
    private final Tree<T>[]    children;
    private final Input.Region location;
    private final boolean      lift;

    public T                 head()        { return head; }

    private Tree<T> lifted() { return children[children.length-1]; }

    public int               numChildren() {
        if (lift && children.length > 0)
            return (children.length-1)+lifted().numChildren();
        return children.length;
    }
    public Iterable<Tree<T>> children()    { return this; }
    public Iterator<Tree<T>> iterator()    {
        if (lift)
            return new ConcatenateIterator(new ArrayIterator(children, 0, children.length-1),
                                           children[children.length-1].iterator());
        return new ArrayIterator(children);
    }
    public Tree<T>           child(int i)  {
        if (lift && i >= children.length-1)
            return children[children.length-1].child(i-(children.length-1));
        return children[i];
    }

    public Input.Region    getRegion() { return location; }

    public Tree(Input.Region loc, T head)                   { this(loc, head, null); }
    public Tree(Input.Region loc, T head, Tree<T>[] children) { this(loc, head, children, false); }
    public Tree(Input.Region loc, T head, Tree<T>[] children, boolean lift) {
        this.location = loc;
        this.head = head;
        this.lift = lift;

        Tree<T>[] children2 = children==null ? new Tree[0] : new Tree[children.length];
        if (children != null) System.arraycopy(children, 0, children2, 0, children.length);
        this.children = children2;
    }

    protected String headToString() { return head()==null?null:head().toString(); }
    protected String headToJava()   {
      // FIXME
        if (head()==null) return null;
        if (head() instanceof ToJava) {
            StringBuffer sb = new StringBuffer();
            ((ToJava)head()).toJava(sb);
            return sb.toString();
        }
        return (head()==null?"null":("\""+StringUtil.toJavaString(head().toString())+"\""));
    }
    protected String left()   { return "{"; }
    protected String right()  { return "}"; }
    protected boolean ignoreSingleton() { return false; }

    public GraphViz.Node toGraphViz(GraphViz gv) {
        if (gv.hasNode(this)) return gv.createNode(this);
        GraphViz.Node n = gv.createNode(this);
        n.label = head()==null ? "" : head().toString();
        //n.color = "red";
        for(Tree t : this) n.edge(t, null);
        return n;
    }
    public boolean isTransparent() { return false; }
    public boolean isHidden() { return false; }

    public static interface TreeFunctor<T,R> extends Functor<Iterable<Tree<T>>, R> {
    }

    public static class ArrayBuildingTreeFunctor<T> implements TreeFunctor<T,T[]>, ToJava {

        public void toJava(StringBuffer sb) { sb.append("new Tree.ArrayBuildingTreeFunctor()"); }
        public String toString() { return ""; }
        
        public T[] invoke(Iterable<Tree<T>> t) {
            ArrayList ret = new ArrayList();
            for(Tree tc : t) {
                if (tc.head() != null && tc.head() instanceof Functor)
                    ret.add(((Functor<Iterable<Tree>,Object>)tc.head()).invoke(tc.children()));
                else if (tc.numChildren() == 0)
                    ret.add(tc.head());
                else {
                    System.err.println("FIXME: don't know what to do about " + tc);
                    ret.add(null);
                }
            }
            return (T[])ret.toArray(new Object[0]);
        }
    }

    public static class RawBindingFunctor<T> implements TreeFunctor<T,Object>, ToJava {
        private Binding _binding;
        private String _toString;
        public RawBindingFunctor(String toString, Binding binding) {
            this._binding = binding;
            this._toString = toString;
        }
        public String toString() {
            return _toString;
        }
        public void toJava(StringBuffer sb) {
            sb.append("new Tree.RawBindingFunctor(");
            sb.append("\"");
            sb.append(StringUtil.toJavaString(_toString));
            sb.append("\", ");
            _binding.toJava(sb);
            sb.append(")");
        }
        public Object invoke(Iterable<Tree<T>> t) {
            return _binding.invoke(new Object[] { t });           
        }
    }

    public static class BindingFunctor<T> implements TreeFunctor<T,Object>, ToJava {
        private Binding _binding;
        private String _toString;
        public String toString() {
            return _toString;
        }
        public BindingFunctor(String toString, Binding binding) {
            this._binding = binding;
            this._toString = toString;
        }
        public void toJava(StringBuffer sb) {
            sb.append("new Tree.BindingFunctor(");
            sb.append("\"");
            sb.append(StringUtil.toJavaString(_toString));
            sb.append("\", ");
            _binding.toJava(sb);
            sb.append(")");
        }
        public Object invoke(Iterable<Tree<T>> t) {
            ArrayList ret = new ArrayList();
            for(Tree tc : t) {
                if (tc.head() != null && tc.head() instanceof Functor)
                    ret.add(((Tree.TreeFunctor<Object,Object>)tc.head()).invoke(tc.children()));
                else if (tc.numChildren() == 0)
                    ret.add(tc.head());
                else {
                    System.err.println("FIXME: don't know what to do about " + tc);
                    ret.add(null);
                }
            }
            Object[] o = (Object[])ret.toArray(new Object[0]);
            return _binding.invoke(o);
        }
    }

}
